'''
These are the things that are used when you `cytoplasm serve`.
'''

import os
import sys
import cytoplasm


# make this work in either Python 2.x or 3.x
if sys.version_info.major >= 3:
    from http.server import SimpleHTTPRequestHandler, HTTPServer
else:
    from SimpleHTTPServer import SimpleHTTPRequestHandler
    from BaseHTTPServer import HTTPServer


# keep track of when things were last built in this global variable
most_recent_time = 0
# keep track of the site in this global variable
site = None


def serve(port, rebuild, event=None):
    "Serve the Cytoplasm site."
    # keep track of the most recently modified time in global variable
    # most_recent_time
    global most_recent_time
    global site
    # create a site and rebuild it first.
    site = cytoplasm.Site(".")
    site.build()
    # set the most recent time.
    most_recent_time = most_recent()
    # change to the build directory, where things are to be served from.
    os.chdir(site.config.build_dir)
    # use either SimpleHTTPRequestHandler or RebuildHandler, depending on
    # whether rebuild is True.
    if rebuild:
        handler = RebuildHandler
    else:
        handler = SimpleHTTPRequestHandler
    # make a server with the handler and the port
    httpd = HTTPServer(('', port), handler)
    # serve!
    httpd.serve_forever()


def most_recent():
    """Determine the most recent modified time in the source directory,
    ignoring dotfiles and _build.
    """
    directory = site.source
    build_dir = site.config.build_dir
    # get the candidate files:
    files = [f for f in os.listdir(directory) if f != build_dir and not
            f.startswith(".")]
    # get each of their times
    times = [os.stat(os.path.join(directory, f)).st_mtime for f in files]
    # the highest time here is the most recent; return that.
    return max(times)


class RebuildHandler(SimpleHTTPRequestHandler):
    def handle(self):
        "Handle a request and, if anything has changed, rebuild the site."
        # overwrite the handle method in SimpleHTTPRequestHandler with this.
        # declare most_recent_time global; we'll be changing it later.
        global most_recent_time
        # figure out the most recent time edited in the source directory
        new_recent = most_recent()
        # only build the site if the new most recent is more recent than the
        # old one, i.e. if one or more of the files has been edited.
        if new_recent > most_recent_time:
            # update most_recent_time
            most_recent_time = new_recent
            # Build the site from the source directory
            print("Rebuilding your Cytoplasm site...")
            site.build()
        # Call SimpleHTTPRequestHandler.handle(), so it can do stuff there too.
        SimpleHTTPRequestHandler.handle(self)



'''
These are some utilites used when writing and handling interpreters.
'''

import shutil
from cytoplasm.errors import InterpreterError


def interpreted_filename(file, site):
    "Intelligently return the interpreted filename of a file."
    # get the last suffix:
    suffix = file.split(".")[-1]
    # if the suffix is in interpreters.keys() the destination is everything
    # but that last suffix.
    if suffix in site.config.interpreters.keys():
        return ".".join(file.split(".")[:-1])
    # otherwise, it's the whole thing.
    else:
        return file


def SaveReturned(fn):
    '''Some potential interpreters, like Mako, don't give you an easy way to
    save to a destination. In these cases, simply use this function as a
    decorator.
    '''
    def InterpreterWithSave(source, destination, **kwargs):
        # destination should _always_ be a file or file-like object.
        # pass **kwargs to this function.
        destination.write(fn(source, **kwargs))

    return InterpreterWithSave


def default_interpreter(source, destination, **kwargs):
    "Copy from source to destination. Use this if no other interpreters match."
    # destination and source are always file objects, so you can copyfileobj.
    shutil.copyfileobj(source, destination)


def interpret(source, destination, site, **kwargs):
    """Interpret a file to a destination with an interpreter according to
    its suffix.
    """
    # open the destination and write to it.
    with open(destination, "w") as destination_file:
        # and, since we now have a file-like object, use interpret_to_filelike
        interpret_to_filelike(source, destination_file, site, **kwargs)


def interpret_to_filelike(source, destination, site, **kwargs):
    """Interpret a file to a file-like object with an interpreter according
    to its suffix."""
    # figure out the last suffix of the file, to use to determine which
    # interpreter to use
    ending = source.split(".")[-1]
    # and then interpret, based on the ending of the file!
    with open(source, "r") as f:
        interpret_filelike(f, destination, site, ending, **kwargs)


def interpret_filelike(source, destination, site, suffix, **kwargs):
    """Given some two file-like objects and a suffix, interpret the source
    according to the given suffix and write it to the destination file object.
    """
    # get the list of interpreters from the configuration
    interpreters = site.config.interpreters
    # and then interpret, based on the ending of the file!
    interpreters.get(suffix, default_interpreter)(source, destination, **kwargs)



'''
A unittest test suite for Cytoplasm.
These work on the example sites in `cytoplasm/tests`.
'''

import os
import imp
import shutil
import unittest
import mako
import cytoplasm
from cytoplasm.site import Site

# Figure out where the examples are, in the cytoplasm package.
_, cytoplasm_directory, _ = imp.find_module("cytoplasm")
# this is the directory
examples_directory = os.path.join(cytoplasm_directory, "tests")


class Base(unittest.TestCase):
    """A base for testing all of the example sites. Default to the 'empty'
    example site.
    """
    def setUp(self, directory=os.path.join(examples_directory, "empty")):
        "Figure out the directories and build the site."
        self.directory = directory
        # create a Site object
        self.site = Site(self.directory)
        # get the build directory from the site's configuration
        self.build_dir = os.path.join(self.site.source,
                self.site.config.build_dir)
        # delete the build directory, just in case there's something there.
        self.delete_build_dir()
        # and, finally, build the site
        self.site.build()

    def delete_build_dir(self):
        "Delete the build directory."
        if os.path.exists(self.build_dir):
            shutil.rmtree(self.build_dir)

    def tearDown(self):
        "Delete the build directory after all the tests are done."
        self.delete_build_dir()

    def test_basic(self):
        "Check for basic decency in a built cytoplasm site."
        # check that the build directory was created
        self.assertTrue(os.path.exists(self.build_dir))
        # check that the build directory is, in fact, a directory
        self.assertTrue(os.path.isdir(self.build_dir))

    def test_copy_html(self):
        """Test whether building a cytoplasm site correctly copies over the
        uninterpreted files."""
        # a filter to tell whether files are html files and are not
        # configuration files.
        filter = lambda x: x.endswith(".html") and not x.startswith("_")
        # the html files in the source directory:
        in_source_dir = [file for file in os.listdir(self.directory) if
                filter(file)]
        # the html files in the build dir:
        in_build_dir = [file for file in os.listdir(self.build_dir) if
                filter(file)]
        # If nothing bad has happenned, everything in in_source_dir should be
        # in in_build_dir
        self.assertTrue(set(in_source_dir) <= set(in_build_dir))
        # Furthermore, the contents of each of these files should be the same.
        for source_path in in_source_dir:
            # open each of these files from their respective directories
            source_file = open(os.path.join(self.directory, source_path), "rb")
            built_file = open(os.path.join(self.build_dir, source_path), "rb")
            # assert that their contents are the same
            self.assertEqual(source_file.read(), built_file.read())
            # and then close each file
            source_file.close()
            built_file.close()


class TestSomeHTML(Base):
    "Test the somehtml example site."
    def setUp(self):
        Base.setUp(self, os.path.join(examples_directory, "somehtml"))

    def test_interpreter(self):
        "Test that the interpreters basically work."
        # I'm not going to do any content testing, because that's mako's job.
        # Still, I'm going to test the files are copied over and interpreted.
        # for each .mako file in the source directory.
        for file in [file for file in os.listdir(self.directory) if
                file.endswith(".mako")]:
            # the corresponding file will be the same, but without the last
            # dotted section.
            corresponding = ".".join(file.split(".")[:-1])
            # assert that there exists a corresponding file in the build
            # directory:
            self.assertTrue(os.path.exists(os.path.join(self.build_dir,
                                                        corresponding)))


class TestControllers(Base):
    "Test the controllers example site."
    def setUp(self):
        Base.setUp(self, os.path.join(examples_directory, "controllers"))

    def test_basic_controller(self):
        # for each controller configured:
        for _, [source_dir, build_dir] in self.site.config.controllers:
            # make sure the build directories were created.
            self.assertTrue(os.path.exists(os.path.join(self.directory,
                                                        build_dir)))

    def test_copier_controller(self):
        # get the controllers whose names are "copier"
        copiers = [c for c in self.site.config.controllers if
                c[0] == "copier"]
        for _, [source_dir, build_dir] in copiers:
            # for each of the files in the build directory...
            for file in os.listdir(os.path.join(self.directory, source_dir)):
                source_file = open(os.path.join(self.directory, source_dir,
                                                file), "rb")
                built_file = open(os.path.join(self.directory, source_dir,
                                                file), "rb")
                # check that they were copied correctly.
                self.assertEqual(source_file.read(), built_file.read())
                # and then close them.
                source_file.close()
                built_file.close()

if __name__ == '__main__':
    unittest.main()



'''These are the things used in making and using controllers.
'''

import imp
import os


class Controller(object):
    """Controllers take data from some files, do stuff with it,
    and write it to the build directory."""
    def __init__(self, site, data, destination, templates="_templates"):
        # take four arguments: the site, the source directory, the destination
        # directory, and optionally a directory wherein the templates reside.
        # assume these are relative pathnames.
        self.site = site
        self.data_directory = os.path.join(site.source, data)
        self.destination_directory = os.path.join(site.source, destination)
        self.templates_directory = os.path.join(site.source, templates)
        # create the destination directory, if it doesn't exist
        if not os.path.exists(self.destination_directory):
            os.mkdir(self.destination_directory)

    def __call__(self):
        # do whatever needs to be done here...
        pass

    def template(self, name):
        """Given a name, return the name of the file in the templates directory
        that fits it."""
        # list of files that fit:
        files = [f for f in os.listdir(self.templates_directory)
                if f.startswith(name + ".")]
        # return the first file that matches, with the directory prepended.
        return os.path.join(self.templates_directory, files[0])



# -*- Coding: utf-8 -*-

import os
import imp
from cytoplasm import interpreters
from cytoplasm.errors import CytoplasmError


class Site(object):
    "The most basic representation of a cytoplasm site."
    def __init__(self, source_directory):
        """Given a source directory, collect information about the site there,
        including configuration files, controllers, and so on.
        """
        self.source = os.path.abspath(source_directory)
        self.config = self._get_config()
        self.build_dir = os.path.join(self.source, self.config.build_dir)

    def _get_config(self):
        """Get the configuration file named `_config.py` from this site's 
        source directory.
        """
        config_file = os.path.join(self.source, "_config.py")
        if os.path.exists(config_file):
            # load the source file as a module and import it.
            imp.load_source("_config", config_file)
            import _config
            return _config
        else:
            # if there's no configuration file, raise an error.
            raise Cytoplasm_Error("You don't seem to have a configuration"
                    "file at %s" % config_file)

    def build(self, target_directory=None):
        "Optionally given a target directory, build the site."
        # if we weren't given a target directory, build into the target
        # directory specified by the configuration file.
        if target_directory == None:
            target_directory = os.path.join(self.source,
                    self.config.build_dir)
        # Create the build directory, if it doesn't exist
        if not os.path.exists(target_directory):
            os.mkdir(target_directory)
        # and copy things over
        self._copy_over(target_directory)
        # and then call all of the controllers
        for controller, arguments in self.config.controllers:
            # create a controller object of the class returned by
            # `self._get_controller` given the site and the name of the
            # controller.
            controller_class = self._get_controller(controller)
            controller_object = controller_class(self, *arguments)
            # call the controller object.
            controller_object()

    def _get_controller(self, name):
        "Given the name of a controller, return the controller's class."
        controller_dir = os.path.join(self.source, "_controllers")
        module = imp.find_module(name, [controller_dir])
        controller = imp.load_module(name, *module)
        return controller.info["class"]

    def _copy_over(self, target_directory):
        """Copy the files and directories not beginning with '_' in the source
        directory to the build directory. If they match an interpreter,
        interpret them first.
        """
        files = ((f, os.path.join(self.source, f)) for f in 
                os.listdir(self.source) if not f.startswith(("_", ".")))
        for base, full in files:
            if os.path.isfile(full):
                destination = interpreters.interpreted_filename(base, self)
                interpreters.interpret(full, 
                        os.path.join(self.build_dir, destination), self)
            elif os.path.isdir(full):
                destination = os.path.join(self.build_dir, base)
                # if the directory exists in the build directory, delete it
                if os.path.exists(destination):
                    shutil.rmtree(destination)
                # and then copy it completely to the build directory
                shutil.copytree(full, destination)
